from typing import Sequence, Collection
from numba import njit
import numpy as np

@njit
def get_ss_w(sqdistances: np.ndarray, labels: np.ndarray, bc: np.ndarray):
    sum = np.zeros_like(bc, dtype=np.float64)
    for i in range(sqdistances.shape[0]):
        for j in range(sqdistances.shape[0]):
            if labels[i] == labels[j] and i != j:
                sum[labels[i]] += sqdistances[i, j]
    return np.sum(sum/bc/2)


def get_ss_t(sqdistances: np.ndarray):
    sum = np.float64(0)
    for i in range(sqdistances.shape[0]):
        for j in range(sqdistances.shape[0]):
            if i != j:
                sum += sqdistances[i, j]

    return sum/sqdistances.shape[0]/2


def get_f(ss_t: np.float64, ss_w: np.float64, a: int, n: int):
    ss_a = ss_t - ss_w
    f = (ss_a/(a-1))/(ss_w/(n-a))
    return f


def permanova(sqdistances: np.ndarray, fastlabels: np.ndarray):

    bc = np.bincount(fastlabels)
    ss_t = get_ss_t(sqdistances)
    ss_w = get_ss_w(sqdistances, fastlabels, bc)

    f = get_f(ss_t, ss_w, len(bc), len(fastlabels))
    other_fs = []
    for _ in range(1000):
        np.random.shuffle(fastlabels)
        other_fs.append(
            get_f(
                ss_t,
                get_ss_w(sqdistances, fastlabels, bc),
                len(bc),
                len(fastlabels)
            )
        )

    return f, (np.array(other_fs) >= f).mean()



from typing import Literal, Union, SupportsFloat, SupportsIndex

_MetricKind = Literal[
    "braycurtis",
    "canberra",
    "chebychev",
    "chebyshev",
    "cheby",
    "cheb",
    "ch",
    "cityblock",
    "cblock",
    "cb",
    "c",
    "correlation",
    "co",
    "cosine",
    "cos",
    "dice",
    "euclidean",
    "euclid",
    "eu",
    "e",
    "hamming",
    "hamm",
    "ha",
    "h",
    "minkowski",
    "mi",
    "m",
    "pnorm",
    "jaccard",
    "jacc",
    "ja",
    "j",
    "jensenshannon",
    "js",
    "kulczynski1",
    "mahalanobis",
    "mahal",
    "mah",
    "rogerstanimoto",
    "russellrao",
    "seuclidean",
    "se",
    "s",
    "sokalmichener",
    "sokalsneath",
    "sqeuclidean",
    "sqe",
    "sqeuclid",
    "yule",
]
_FloatValue = Union[None, str, bytes, SupportsFloat, SupportsIndex]



from __future__ import annotations
from typing import (
    TypeVar,
    Collection,
    Callable,
    Literal,
    overload,
    TypeGuard,
    Any,
    TYPE_CHECKING,
)
import numpy as np
from numpy.typing import NDArray
import polars as pl
from scipy.spatial.distance import pdist
from ._oxide import permanova

if TYPE_CHECKING:
    from ._scipy_types import _FloatValue, _MetricKind

T = TypeVar("T")


def get_distance_matrix(things: Collection[T], distance: Callable[[T, T], _FloatValue]):
    dists = np.empty((len(things), len(things)), dtype=np.float64)
    for i, a in enumerate(things):
        for j, b in enumerate(things):
            if i != j:
                dists[i, j] = np.float64(distance(a, b))
    return dists


RT = TypeVar("RT")
CT = TypeVar("CT")


# the runtime check will be performed by the package
# am i having too much fun with the type system?
def pinky_promise_guard(
    ct: Callable[[T, T], RT], t: T, check_type: type[CT]
) -> TypeGuard[Callable[[CT, CT], RT]]:
    """
    check if `t:T` is of type `check_type: CT` and if so,
    rely on the fact that the callable accepts [T,T] (same as t!)
    to infer it must also be [CT,CT].
    This narrowing should be automatic but oh well
    """
    return isinstance(t, check_type)


def calculate_distances(
    things: Collection[T] | NDArray[Any],
    distance: Callable[[T, T], _FloatValue] | _MetricKind,
    engine: str,
) -> NDArray[np.floating[Any]]:
    if len(things) < 2:
        raise ValueError("len(things) < 2")

    if engine == "scipy":
        if isinstance(things, np.ndarray) and len(things.shape) == 2:
            if callable(distance) and pinky_promise_guard(
                distance, things[0], NDArray[Any]
            ):
                return pdist(things, distance)
            elif isinstance(distance, str):
                return pdist(things, distance)  # type inference in overload hello?
            else:
                raise ValueError("distance wrong")
        else:
            raise ValueError("things wrong")

    elif engine == "python":
        if isinstance(distance, str):
            raise ValueError("distance wrong")
        return get_distance_matrix(things, distance)

    elif engine == "numba":
        if isinstance(distance, str):
            raise ValueError("distance wrong")
        import numba  # type: ignore[reportMissingTypeStubs]

        return numba.jit(get_distance_matrix)(things, distance)  # type: ignore

    raise ValueError("engine wrong")


@overload
def run(
    things: NDArray[Any],
    distance: _MetricKind,
    labels: Collection[Any],
    engine: Literal["scipy"],
) -> tuple[float, float]:
    ...


@overload
def run(
    things: Collection[T],
    distance: Callable[[T, T], _FloatValue],
    labels: Collection[Any],
    engine: Literal["python", "numba"],
) -> tuple[float, float]:
    ...


def run(
    things: Collection[T] | NDArray[Any],
    distance: Callable[[T, T], _FloatValue] | _MetricKind,
    labels: Collection[Any],
    engine: Literal["scipy", "python", "numba"],
) -> tuple[float, float]:
    dist = calculate_distances(things, distance, engine)
    fastlabels = (
        pl.Series(labels).cast(pl.Categorical()).cast(pl.Int64()).to_numpy().copy()
    )
    return permanova(dist, fastlabels)



from skbio.stats.distance import permanova, DistanceMatrix
from reference import square
import peroxymanova
import numpy as np
from scipy.spatial import distance_matrix
from scipy.stats import f_oneway

size = 10
objects = np.random.random((size, 1))
dist = distance_matrix(objects, objects)
labels = np.random.randint(0, 2, size)
standard = permanova(
    DistanceMatrix(dist),
    labels.copy(),
    permutations=1000
)

python = square.permanova(dist**2, labels.copy())

our = peroxymanova.permanova(dist**2, labels.copy())
anova = f_oneway(*(objects[labels == i] for i in np.unique(labels)))

print(standard)
print(python)
print(our)
print(anova)

assert np.allclose(standard['test statistic'], our[0])
